@@ -393,12 +393,160 @@ class Data:
         self.print_depth = depth
 
 
+class DicomFile(Data):
+    '''
+    Class representing files in DICOM format, with conversion to skrt objects.
+
+    **Methods:**
+    - **_Init__()** : Create instance of DicomFile class.
+    - **get_object()** : 
+    - **get_matched_attributes()** : 
+    - **set_dates_and_times()** : 
+    - **set_referenced_sop_instance_uids** :
+    - **set_slice_thickness** :
+    '''
+
+    def __init__(self, path):
+        self.path = str(path)
+        try:
+            self.ds = pydicom.dcmread(self.path, force=True)
+        except IsADirectoryError:
+            self.ds = None
+        elements = {
+                "study": ["Study", "Content", "InstanceCreation"],
+                "item": ["Instance", "Content", "Series", "InstanceCreation"]
+                }
+        self.set_dates_and_times(elements)
+        self.set_slice_thickness()
+        self.frame_of_reference_uid = getattr(
+                self.ds, "FrameOfReferenceUID", None)
+        self.modality = getattr(self.ds, "Modality", "unknown")
+        if self.modality:
+            self.modality = self.modality.lower()
+        self.series_number = getattr(self.ds, "SeriesNumber", None)
+        self.series_instance_uid = getattr(self.ds, "SeriesInstanceUID", None)
+        self.sop_instance_uid = getattr(self.ds, "SOPInstanceUID", None)
+        self.study_instance_uid = getattr(self.ds, "StudyInstanceUID", None)
+        self.set_referenced_sop_instance_uids()
+
+    def get_object(self, cls, **kwargs):
+
+        if cls.__name__ in ["Dose", "Plan", "StructureSet"]:
+            kwargs["path"] = self.path
+            kwargs["load"] = False
+
+        obj = cls(**kwargs)
+
+        if cls.__name__ == "Study":
+            obj.date = self.study_date
+            obj.time = self.study_time
+            obj.timestamp = self.study_timestamp
+            for item in ["dose", "image", "plan", "structure_set"]:
+                setattr(obj, f"{item}_types", {})
+        else:
+            obj.date = self.item_date
+            obj.time = self.item_time
+            obj.timestamp = self.item_timestamp
+
+        obj.modality = self.modality
+        obj.frame_of_reference_uid = self.frame_of_reference_uid
+        obj.referenced_image_sop_instance_uid = (
+                self.referenced_image_sop_instance_uid)
+        obj.referenced_plan_sop_instance_uid = (
+                self.referenced_plan_sop_instance_uid)
+        obj.referenced_structure_set_sop_instance_uid = (
+                self.referenced_structure_set_sop_instance_uid)
+        obj.series_instance_uid = self.series_instance_uid
+        obj.series_number = self.series_number
+        obj.sop_instance_uid = self.sop_instance_uid
+        obj.study_instance_uid = self.study_instance_uid
+
+        return obj
+
+    def get_matched_attributes(self, others, attributes=None):
+        attributes = attributes or []
+        if isinstance(attributes, str):
+            attributes = [attributes]
+        matches = others or []
+
+        for attribute in attributes:
+            matches = [match for match in matches if
+                    (getattr(match, attribute, None)
+                    == getattr(self, attribute, None))]
+
+        return matches 
+
+    def set_dates_and_times(self, elements):
+
+        measurements = {"Date": "00000000", "Time": "000000"}
+
+        for element, timed_items in elements.items():
+            for measurement, value in measurements.items():
+                attribute = f"{element}_{measurement.lower()}"
+                if self.ds is None:
+                    setattr(self, attribute, None)
+                    continue
+                setattr(self, attribute, value)
+                for timed_item in timed_items:
+                    target = f"{timed_item}{measurement}"
+                    if hasattr(self.ds, target):
+                        target_value = getattr(self.ds, target)
+                        if target_value:
+                            setattr(self, attribute, target_value.split(".")[0])
+                            break
+
+            if self.ds is None:
+                timestamp = None
+            else:
+                date = getattr(self, f"{element}_date")
+                time = getattr(self, f"{element}_time")
+                setattr(self, f"{element}_timestamp", f"{date}_{time}")
+
+    def set_referenced_sop_instance_uids(self):
+        try:
+            uid = self.ds.ReferencedImageSequence[0].ReferencedSOPInstanceUID
+        except (IndexError, AttributeError):
+            uid = None
+
+        if uid is None:
+            try:
+                uid = (self.ds.ROIContourSequence[0].ContourSequence[0]
+                        .ContourImageSequence[0].ReferencedSOPInstanceUID)
+            except (IndexError, AttributeError):
+                pass
+
+        self.referenced_image_sop_instance_uid = uid
+
+        try:
+            uid = (self.ds.ReferencedStructureSetSequence[0]
+                    .ReferencedSOPInstanceUID)
+        except (IndexError, AttributeError):
+            None
+
+        self.referenced_structure_set_sop_instance_uid = uid
+
+        try:
+            uid = (self.ds.ReferencedRTPlanSequence[0]
+                    .ReferencedSOPInstanceUID)
+        except (IndexError, AttributeError):
+            uid = None
+
+        self.referenced_plan_sop_instance_uid = uid
+
+    def set_slice_thickness(self):
+        self.slice_thickness = ""
+        if hasattr(self.ds, "SliceThickness"):
+            slice_thickness = self.ds.SliceThickness
+            if slice_thickness:
+                self.slice_thickness = f"{self.ds.SliceThickness : .3f}"
+
+
 class PathData(Data):
     """Data with an associated path or directory; has the ability to
     extract a list of dated objects from within this directory."""
 
     def __init__(self, path=""):
-        self.path = fullpath(path)
+        self.path = fullpath(str(path))
         self.subdir = ""
 
     def create_objects(
@@ -532,6 +680,19 @@ class Dated(PathData):
                 return False
         return True
 
+    def get_matched_timestamps(self, others=None, delta_seconds=120):
+        others = others or []
+        matches = []
+        for other in others:
+            try:
+                delta_time = abs(float(other.time) - float(self.time))
+            except NoneType:
+                delta_time = 0
+            if other.date == self.date and delta_time <= delta_seconds:
+                matches.append(other)
+
+        return matches
+
     def __eq__(self, other):
         return self.date == other.date and self.time == other.time
 
@@ -667,7 +828,7 @@ def alphanumeric(in_str: str = "") -> List[str]:
     return elements
 
 
-def fullpath(path: str = "") -> str:
+def fullpath(path=""):
     """Evaluate full path, expanding '~', environment variables, and
     symbolic links."""
 
@@ -892,6 +1053,10 @@ def get_time_separated_objects(objs, min_delta=4, unit='hour',
         the most recent if True, or the least recent otherwise. 
     '''
 
+    # Deal with cases where the number of input objects is 0 or 1.
+    if len(objs) <= 1:
+        return objs
+
     # Ensure that min_delta is a pandas Timedelta object.
     if not isinstance(min_delta, pd.Timedelta):
         min_delta = pd.Timedelta(min_delta, unit)
@@ -929,3 +1094,121 @@ def get_time_separated_objects(objs, min_delta=4, unit='hour',
         time_separated_objs.reverse()
 
     return time_separated_objs
+
+def get_interval_in_days(timestamp1, timestamp2):
+    '''
+    Return interval in days between two timestamps.
+
+    **Parameters:**
+
+    timestamp1 : pandas.Timestamp
+        Timestamp corresponding to start of interval.
+
+    timestamp2 : pandas.Timestamp
+        Timestamp corresponding to end of interval.
+    '''
+    if (isinstance(timestamp1, pd.Timestamp) and
+            isinstance(timestamp2, pd.Timestamp)):
+        interval = (timestamp2 - timestamp1).total_seconds()
+        interval /= pd.Timedelta('1d').total_seconds()
+    else:
+        interval = None
+    return interval
+
+def get_interval_in_whole_days(timestamp1, timestamp2):
+    '''
+    Return interval in whole days between two timestamps.
+    
+    **Parameters:**
+
+    timestamp1 : pandas.Timestamp
+        Timestamp corresponding to start of interval.
+
+    timestamp2 : pandas.Timestamp
+        Timestamp corresponding to end of interval.
+    '''
+    if (isinstance(timestamp1, pd.Timestamp) and
+            isinstance(timestamp2, pd.Timestamp)):
+        interval = ((timestamp2.floor('1d') - timestamp1.floor('1d'))
+                .total_seconds())
+        interval /= pd.Timedelta('1d').total_seconds()
+        interval = round(interval)
+    else:
+        interval = None
+    return interval
+
+def year_fraction(timestamp):
+    '''
+    Convert from timestamp to year, including fractional part.
+    
+    **Parameter:**
+
+    timestamp : pandas.Timestamp
+        Timestamp to be converted.
+    '''
+    if isinstance(timestamp, pd.Timestamp):
+        # Determine year.
+        year = timestamp.year
+
+        # Determine seconds in year (different for leap year and non-leap year).
+        year_start = pd.Timestamp(f'{year}0101')
+        next_year_start = pd.Timestamp(f'{year + 1}0101')
+        seconds_in_year = (next_year_start - year_start).total_seconds()
+
+        # Determine seconds elapsed so far in year.
+        seconds_to_date = (timestamp - year_start).total_seconds()
+
+        # Add year and fractional part.
+        year_fraction = year + seconds_to_date / seconds_in_year
+    else:
+        year_fraction = None
+    return year_fraction
+
+def get_uid_without_slice(uid):
+    return ".".join(uid.split(".")[:-1])
+
+def get_sequence_value(ds=None, sequence=None, tag=None):
+    value = None
+    sequence_data = getattr(ds, sequence, None)
+    if None not in [ds, sequence, tag]:
+        sequence_data = getattr(ds, sequence, None)
+        if sequence_data:
+            value = getattr(sequence_data[-1], tag, None)
+    return value
+
+def get_referenced_image(referrer=None, image_types=None):
+    image_types = image_types or {}
+    image = None
+    for modality, images in image_types.items():
+        image = get_referenced_object(referrer, images,
+                "referenced_image_sop_instance_uid", True)
+        if image:
+            break
+    if image is None:
+        for modality, images in image_types.items():
+            matched_attributes = DicomFile.get_matched_attributes(
+                    referrer, images, "frame_of_reference_uid")
+            if matched_attributes:
+                image = matched_attributes[0]
+
+    return image
+
+def get_referenced_object(referrer, others, tag, omit_slice=False):
+    referenced_object = None
+    uid1 = getattr(referrer, tag, None)
+    if uid1:
+        if omit_slice:
+            uid1 = get_uid_without_slice(uid1)
+
+        for other in others:
+            if omit_slice:
+                uid2 = get_uid_without_slice(other.sop_instance_uid)
+            else:
+                uid2 = other.sop_instance_uid
+
+            if uid1 == uid2:
+                referenced_object = other
+                break
+
+    return referenced_object
+
