@@ -14,6 +14,9 @@ from skrt.image import Image
 from skrt.structures import StructureSet
 from skrt.dose import Dose, Plan
 
+# Default mode for reading patient data.
+skrt.core.Defaults({"unsorted_dicom": False})
+
 
 class Study(skrt.core.Archive):
     """Class representing a single study; can contain images, dose fields,
@@ -769,7 +772,14 @@ class Patient(skrt.core.PathData):
     a patient ID, and whose subdirectories contain studies.
     """
 
-    def __init__(self, path="", exclude=["logfiles"]):
+    def __init__(self, path=None, exclude=None, unsorted_dicom=None,
+            id_mappings=None):
+
+        # Initialise parameters.
+        path = path or ""
+        exclude = exclude or ["logfiles"]
+        unsorted_dicom = unsorted_dicom or skrt.core.Defaults().unsorted_dicom
+        id_mappings = id_mappings or {}
 
         # Record start time
         tic = timeit.default_timer()
@@ -778,19 +788,24 @@ class Patient(skrt.core.PathData):
         if path is None:
             path = os.getcwd()
         self.path = skrt.core.fullpath(str(path))
-        self.id = os.path.basename(self.path)
+        patient_id = os.path.basename(self.path)
+        self.id = id_mappings.get(patient_id, patient_id)
 
         # Find studies
-        self.studies = self.create_objects(dtype=Study)
-        if not self.studies:
-            if os.path.isdir(self.path):
-                if os.access(self.path, os.R_OK):
-                    subdirs = sorted(os.listdir(self.path))
-                    for subdir in subdirs:
-                        if subdir not in exclude:
-                            self.studies.extend(
-                                self.create_objects(dtype=Study, subdir=subdir)
-                            )
+        if unsorted_dicom:
+            self.sort_dicom()
+        else:
+            self.studies = self.create_objects(dtype=Study)
+            if not self.studies:
+                if os.path.isdir(self.path):
+                    if os.access(self.path, os.R_OK):
+                        subdirs = sorted(os.listdir(self.path))
+                        for subdir in subdirs:
+                            if subdir not in exclude:
+                                self.studies.extend(
+                                        self.create_objects(dtype=Study,
+                                            subdir=subdir)
+                                        )
         
         # Add to data objects a reference to the patient object.
         self.link_study_data_to_patient()
@@ -799,6 +814,129 @@ class Patient(skrt.core.PathData):
         toc = timeit.default_timer()
         self._init_time = (toc - tic)
 
+    def sort_dicom(self):
+        data_types = {
+                "rtplan": "plan",
+                "rtstruct": "structure_set",
+                "rtdose": "dose",
+                }
+        data_classes = {
+                "dose": Dose,
+                "image": Image,
+                "plan": Plan,
+                "structure_set": StructureSet,
+                }
+        self.studies = []
+        patient_path = Path(self.path)
+        file_paths = sorted(patient_path.glob("**/*"))
+        for file_path in file_paths:
+            dcm = skrt.core.DicomFile(file_path)
+            if dcm.ds:
+                matched_attributes = dcm.get_matched_attributes(self.studies,
+                        "study_instance_uid")
+                if matched_attributes:
+                    study = matched_attributes[0]
+                else:
+                    study = dcm.get_object(Study)
+                    self.studies.append(study)
+
+                dstring = data_types.get(dcm.modality, "image")
+                dtypes = getattr(study, f"{dstring}_types")
+                if not dcm.modality in dtypes:
+                    dtypes[dcm.modality] = []
+                if "image" == dstring:
+                    matched_attributes = dcm.get_matched_attributes(
+                            dtypes[dcm.modality],
+                            "series_instance_uid"
+                            )
+                    if matched_attributes:
+                        matched_attributes[0].dicom_paths.append(dcm.path)
+                    else:
+                        dcm.dicom_paths = [dcm.path]
+                        dtypes[dcm.modality].append(dcm)
+                else:
+                    dtypes[dcm.modality].append(dcm)
+
+        for study in self.studies:
+            for dstring, dclass in data_classes.items():
+                dtypes = getattr(study, f"{dstring}_types")
+                for modality in dtypes:
+                    datastore = f"{modality}_{dstring}s"
+                    setattr(study, datastore, [])
+                    for idx, dcm in enumerate(dtypes[modality]):
+                        obj = dcm.get_object(dclass)
+                        if "image" == dstring:
+                            obj.dicom_paths = list(dcm.dicom_paths)
+                            dtypes[modality][idx] = obj
+                        getattr(study, datastore).append(obj)
+
+        for study in self.studies:
+            if study.plan_types and study.structure_set_types:
+                for plan in study.rtplan_plans:
+                    ss = skrt.core.get_referenced_object(
+                            plan, study.rtstruct_structure_sets,
+                            "referenced_structure_set_sop_instance_uid")
+                    if ss:
+                        plan.set_structure_set(ss)
+
+            if study.plan_types and study.dose_types:
+                for dose in study.rtdose_doses:
+                    plan = skrt.core.get_referenced_object(
+                            dose, study.rtplan_plans,
+                            "referenced_plan_sop_instance_uid")
+                    if plan:
+                        dose.set_plan(plan)
+
+        for study in self.studies:
+            for dcm_modality, dstring in data_types.items():
+                dtypes = getattr(study, f"{dstring}_types")
+                dcm_datastore = f"{dcm_modality}_{dstring}s"
+                orphan_objs = []
+                for obj in getattr(study, dcm_datastore, []):
+                    image = skrt.core.get_referenced_image(
+                            obj, study.image_types)
+                    obj.set_image(image)
+
+            if study.plan_types and study.structure_set_types:
+                for plan in study.rtplan_plans:
+                    if not plan.image:
+                        ss = skrt.core.get_referenced_object(
+                                plan, study.rtstruct_structure_sets,
+                                "referenced_structure_set_sop_instance_uid")
+
+                        image = ss.image if ss is not None else None
+                        plan.set_image(image)
+
+                for ss in study.rtstruct_structure_sets:
+                    if not ss.image:
+                        for plan in ss.plans:
+                            if plan.image:
+                                ss.set_image(plan.image)
+                                break
+
+            for dcm_modality, dstring in data_types.items():
+                dtypes = getattr(study, f"{dstring}_types")
+                dcm_datastore = f"{dcm_modality}_{dstring}s"
+                orphan_objs = []
+                for obj in getattr(study, dcm_datastore, []):
+                    if obj.image:
+                        datastore = f"{image.modality}_{dstring}s"
+                        if not image.modality in dtypes:
+                            dtypes[image.modality] = []
+                            setattr(study, datastore, [])
+                        dtypes[image.modality].append(obj)
+                        getattr(study, datastore).append(obj)
+                    else:
+                        orphan_objs.append(obj)
+                if orphan_objs:
+                    setattr(study, dcm_datastore, orphan_objs)
+                    dtypes[dcm_modality] = orphan_objs
+                else:
+                    if hasattr(study, dcm_datastore):
+                        delattr(study, dcm_datastore)
+                    if dcm_modality in dtypes:
+                        del dtypes[dcm_modality]
+
     def link_study_data_to_patient(self):
         '''Add to data objects a reference to the associated patient object.'''
 
@@ -909,6 +1047,7 @@ class Patient(skrt.core.PathData):
             for obj_type in sorted(obj_types):
                 if not obj_type in all_types:
                     all_types.append(obj_type)
+
         all_types.sort()
         return all_types
 
@@ -1062,7 +1201,7 @@ class Patient(skrt.core.PathData):
         info['plan_fraction'] = None
         info['plan_target_dose'] = None
         for study in self.studies:
-            if hasattr(study, 'plan_types'):
+            if hasattr(study, 'plan_types') and study.plan_types:
                 plan_type = sorted(list(study.plan_types.keys()))[0]
                 plan = study.plan_types[plan_type][0]
                 plan.load()
@@ -1129,12 +1268,24 @@ class Patient(skrt.core.PathData):
             info['plan_image_time'] = plan_images[0].get_pandas_timestamp()
         else:
             info['plan_image_time'] = None
+        info['plan_image_year'] = skrt.core.year_fraction(
+                info['plan_image_time']) 
 
         # Store time of plan creation.
         if info['plan_fraction'] is not None:
             info['plan_time'] = plan.get_pandas_timestamp()
         else:
             info['plan_time'] = None
+        info['plan_year'] = skrt.core.year_fraction(
+                info['plan_time']) 
+
+        # Store number of days from planning image to plan creation.
+        info['days_plan_image_to_plan'] = (
+                skrt.core.get_interval_in_days(
+                info['plan_image_time'], info['plan_time']))
+        info['whole_days_plan_image_to_plan'] = (
+                skrt.core.get_interval_in_whole_days(
+                info['plan_image_time'], info['plan_time']))
 
         # Store times of first and last treatment images.
         if treatment_images:
@@ -1143,6 +1294,40 @@ class Patient(skrt.core.PathData):
         else:
             info['treatment_start'] = None
             info['treatment_end'] = None
+        info['treatment_start_year'] = skrt.core.year_fraction(
+                info['treatment_start']) 
+        info['treatment_end_year'] = skrt.core.year_fraction(
+                info['treatment_end']) 
+
+        # Store treatment duration.
+        info['days_treatment'] = skrt.core.get_interval_in_days(
+                info['treatment_start'], info['treatment_end'])
+        info['whole_days_treatment'] = skrt.core.get_interval_in_whole_days(
+                info['treatment_start'], info['treatment_end'])
+
+        # Store number of days from planning to treatment start.
+        info['days_plan_to_treatment'] = (
+                skrt.core.get_interval_in_days(
+                info['plan_time'], info['treatment_start']))
+        info['whole_days_plan_to_treatment'] = (
+                skrt.core.get_interval_in_whole_days(
+                info['plan_time'], info['treatment_start']))
+
+        # Store number of days from planning image to treatment start.
+        info['days_plan_image_to_treatment'] = (
+                skrt.core.get_interval_in_days(
+                info['plan_image_time'], info['treatment_start']))
+        info['whole_days_plan_image_to_treatment'] = (
+                skrt.core.get_interval_in_whole_days(
+                info['plan_image_time'], info['treatment_start']))
+
+        # Check that intervals are consistent.
+        if (info['whole_days_plan_image_to_plan'] and
+                info['whole_days_plan_to_treatment'] and
+                info['whole_days_plan_image_to_treatment']):
+            assert (info['whole_days_plan_image_to_treatment'] ==
+                    info['whole_days_plan_image_to_plan'] +
+                    info['whole_days_plan_to_treatment'])
 
         # Filter to have images separated by a minimum amount of time.
         time_separated_plan_images = skrt.core.get_time_separated_objects(
@@ -1193,20 +1378,38 @@ class Patient(skrt.core.PathData):
             images = self.combined_objs(image_label)
             time_separated_images = skrt.core.get_time_separated_objects(
                     images, min_delta=min_delta, unit=unit)
-            ref_time = None
+
+            time_last = None
+            time_zero = time_separated_images[0].get_pandas_timestamp()
+
             for image in time_separated_images:
                 info = {}
                 info['id'] = self.id
                 info['modality'] = image_type
                 info['timestamp'] = image.get_pandas_timestamp()
-                info['day'] = info['timestamp'].isoweekday()
-                if ref_time is not None:
-                    info['time_delta'] = info['timestamp'] - ref_time
-                    info['days_delta'] = info['time_delta'].round('1d')
-                else:
-                    info['time_delta'] = None
-                    info['days_delta'] = None
-                ref_time = info['timestamp']
+                info['day'] = None
+                info['time_delta'] = None
+                info['days_delta'] = None
+                info['whole_days_delta'] = None
+                info['days_total'] = None
+                info['whole_days_total'] = None
+                if info['timestamp']:
+                    info['day'] = info['timestamp'].isoweekday()
+                    if time_last is not None:
+                        info['time_delta'] = info['timestamp'] - time_last
+                    else:
+                        info['time_delta'] = None
+                    info['days_delta'] = skrt.core.get_interval_in_days(
+                            time_last, info['timestamp'])
+                    info['whole_days_delta'] = (
+                            skrt.core.get_interval_in_whole_days(
+                                time_last, info['timestamp']))
+                    time_last = info['timestamp']
+                    info['days_total'] = skrt.core.get_interval_in_days(
+                            time_zero, info['timestamp'])
+                    info['whole_days_total'] = (
+                            skrt.core.get_interval_in_whole_days(
+                                time_zero, info['timestamp']))
                 all_info.append(info)
 
         return (pd.DataFrame(all_info) if df else all_info)
